# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, Set

from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.graph.memory_graph.memory_element import MemoryElements


class Cut:
    """
    A Cut object that contains a set of ordered nodes and their memory elements.
    """

    def __init__(self, op_order: List[BaseNode], op_record: Set[BaseNode], mem_elements: MemoryElements):
        """
        Args:
            op_order: A list of the cut's nodes (model layers), ordered by their addition to the cut (first-to-last).
            op_record: A (unordered) set of the nodes in the cut.
            mem_elements: MemoryElements object which represents the activation tensors of the cut's nodes.
        """

        self.op_order = op_order
        self.op_record = op_record
        self.mem_elements = mem_elements

    def memory_size(self) -> float:
        """
        Returns: The total memory size of the cut.
        """

        return self.mem_elements.total_size

    def get_record_names(self) -> Set[str]:
        """
        Builds a set of the cut nodes' names.

        Returns: a set with the nodes' names.
        """

        return {op.name for op in self.op_record}

    def __eq__(self, other) -> bool:
        """
        Overrides the class equality method.
        Two Cuts are equal if they contain the same memory elements.

        Args:
            other: An object to compare the current object to.

        Returns: True if the two objects are equal. False otherwise.

        """
        if isinstance(other, Cut):
            return self.mem_elements == other.mem_elements
        return False

    def __hash__(self):
        return hash((frozenset(self.op_order), frozenset(self.op_record), self.mem_elements))


# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from typing import Any, Callable

import torch
from tqdm import tqdm

from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core import common
from model_compression_toolkit.core.pytorch.constants import GAMMA, BETA, MOVING_MEAN, MOVING_VARIANCE
from model_compression_toolkit.core.pytorch.utils import set_model, to_torch_tensor


def pytorch_apply_second_moment_correction(quantized_model: Any,
                                           core_config: CoreConfig,
                                           representative_data_gen: Callable,
                                           graph: common.Graph):
    """
    Apply second moment statistics correction to graph.

    Args:
        quantized_model: Framework's model to apply second moment correction on.
        core_config: QuantizationConfig of how the model should be quantized.
        representative_data_gen: Dataset to use for retrieving images for the models inputs.
        graph: Graph to update the parameters after the second moment correction.

    Returns:
        A function that applies second moment correction.
    """
    model = copy.deepcopy(quantized_model)
    set_model(model)

    # Move every BN to train mode
    for name, module in model.named_modules():
        if len(graph.find_node_by_name(name)) > 0:
            node = graph.find_node_by_name(name)[0]
            if isinstance(module, torch.nn.BatchNorm2d) and node.final_weights_quantization_cfg\
                    .weights_second_moment_correction:
                module.train()

    with torch.no_grad():
        for data in tqdm(representative_data_gen()):
            model(*to_torch_tensor(data))

    set_model(model)

    # Move every BN to eval mode and update the corresponding BN node params in the graph
    for name, module in model.named_modules():
        if len(graph.find_node_by_name(name)) > 0:
            node = graph.find_node_by_name(name)[0]
            if isinstance(module, torch.nn.BatchNorm2d) and node.final_weights_quantization_cfg\
                    .weights_second_moment_correction:
                module.eval()
                bn_node_weights = {GAMMA: module.weight.detach().cpu().numpy(),
                                   BETA: module.bias.detach().cpu().numpy(),
                                   MOVING_MEAN: module.running_mean.detach().cpu().numpy(),
                                   MOVING_VARIANCE: module.running_var.detach().cpu().numpy()}
                node.weights = copy.deepcopy(bn_node_weights)

    return graph




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable, Tuple
from packaging import version

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import TENSORFLOW
from model_compression_toolkit.core.common.user_info import UserInformation
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfigV2
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPI
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfigV2
from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core.runner import core_runner, _init_tensorboard_writer
from model_compression_toolkit.gptq.runner import gptq_runner
from model_compression_toolkit.core.exporter import export_model
from model_compression_toolkit.core.analyzer import analyzer_model_quantization
from model_compression_toolkit.core.common.target_platform.targetplatform2framework import TargetPlatformCapabilities

LR_DEFAULT = 0.15
LR_REST_DEFAULT = 1e-4
LR_BIAS_DEFAULT = 1e-4
LR_QUANTIZATION_PARAM_DEFAULT = 1e-3
GPTQ_MOMENTUM = 0.9

if common.constants.FOUND_TF:
    import tensorflow as tf
    from model_compression_toolkit.core.keras.default_framework_info import DEFAULT_KERAS_INFO
    from model_compression_toolkit.core.keras.keras_implementation import KerasImplementation
    from model_compression_toolkit.core.keras.keras_model_validation import KerasModelValidation
    from tensorflow.keras.models import Model
    from model_compression_toolkit.gptq.keras.gptq_loss import GPTQMultipleTensorsLoss
    from model_compression_toolkit.core.keras.constants import DEFAULT_TP_MODEL
    from model_compression_toolkit.exporter.model_wrapper import get_exportable_keras_model
    from model_compression_toolkit import get_target_platform_capabilities

    # As from TF2.9 optimizers package is changed
    if version.parse(tf.__version__) < version.parse("2.9"):
        from keras.optimizer_v2.optimizer_v2 import OptimizerV2
    else:
        from keras.optimizers.optimizer_v2.optimizer_v2 import OptimizerV2

    DEFAULT_KERAS_TPC = get_target_platform_capabilities(TENSORFLOW, DEFAULT_TP_MODEL)


    def get_keras_gptq_config(n_epochs: int,
                              optimizer: OptimizerV2 = tf.keras.optimizers.Adam(learning_rate=LR_DEFAULT),
                              optimizer_rest: OptimizerV2 = tf.keras.optimizers.Adam(learning_rate=LR_REST_DEFAULT),
                              loss: Callable = GPTQMultipleTensorsLoss(),
                              log_function: Callable = None) -> GradientPTQConfigV2:
        """
        Create a GradientPTQConfigV2 instance for Keras models.

        args:
            n_epochs (int): Number of epochs for running the representative dataset for fine-tuning.
            optimizer (OptimizerV2): Keras optimizer to use for fine-tuning for auxiliry variable with a default learning rate set to 0.2.
            optimizer_rest (OptimizerV2): Keras optimizer to use for fine-tuning of the bias variable.
            loss (Callable): loss to use during fine-tuning. should accept 4 lists of tensors. 1st list of quantized tensors, the 2nd list is the float tensors, the 3rd is a list of quantized weights and the 4th is a list of float weights.
            log_function (Callable): Function to log information about the gptq process.

        returns:
            a GradientPTQConfigV2 object to use when fine-tuning the quantized model using gptq.

        Examples:

            Import MCT and TensorFlow:

            >>> import model_compression_toolkit as mct
            >>> import tensorflow as tf

            Create a GradientPTQConfigV2 to run for 5 epochs:

            >>> gptq_conf = mct.get_keras_gptq_config(n_epochs=5)

            Other Tensorflow optimizers can be passed:

            >>> gptq_conf = mct.get_keras_gptq_config(n_epochs=3, optimizer=tf.keras.optimizers.Nadam())

            The configuration can be passed to :func:`~model_compression_toolkit.keras_post_training_quantization` in order to quantize a keras model using gptq.

        """
        bias_optimizer = tf.keras.optimizers.SGD(learning_rate=LR_BIAS_DEFAULT, momentum=GPTQ_MOMENTUM)
        optimizer_quantization_parameter = tf.keras.optimizers.SGD(learning_rate=LR_QUANTIZATION_PARAM_DEFAULT, momentum=GPTQ_MOMENTUM)
        return GradientPTQConfigV2(n_epochs,
                                   optimizer,
                                   optimizer_rest=optimizer_rest,
                                   loss=loss,
                                   log_function=log_function,
                                   train_bias=True,
                                   quantization_parameters_learning=True,
                                   optimizer_bias=bias_optimizer,
                                   optimizer_quantization_parameter=optimizer_quantization_parameter)


    def keras_gradient_post_training_quantization_experimental(in_model: Model,
                                                               representative_data_gen: Callable,
                                                               gptq_config: GradientPTQConfigV2,
                                                               gptq_representative_data_gen: Callable = None,
                                                               target_kpi: KPI = None,
                                                               core_config: CoreConfig = CoreConfig(),
                                                               fw_info: FrameworkInfo = DEFAULT_KERAS_INFO,
                                                               target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_KERAS_TPC,
                                                               new_experimental_exporter: bool = False) -> \
    Tuple[Model, UserInformation]:
        """
        Quantize a trained Keras model using post-training quantization. The model is quantized using a
        symmetric constraint quantization thresholds (power of two).
        The model is first optimized using several transformations (e.g. BatchNormalization folding to
        preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
        being collected for each layer's output (and input, depends on the quantization configuration).
        For each possible bit width (per layer) a threshold is then being calculated using the collected
        statistics. Then, if given a mixed precision config in the core_config, using an ILP solver we find
        a mixed-precision configuration, and set a bit-width for each layer. The model is then quantized
        (both coefficients and activations by default).
        In order to limit the maximal model's size, a target KPI need to be passed after weights_memory
        is set (in bytes).
        Then, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized models, and minimizing the observed
        loss.

        Args:
            in_model (Model): Keras model to quantize.
            representative_data_gen (Callable): Dataset used for calibration.
            gptq_config (GradientPTQConfigV2): Configuration for using gptq (e.g. optimizer).
            gptq_representative_data_gen (Callable): Dataset used for GPTQ training. If None defaults to representative_data_gen
            target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
            core_config (CoreConfig): Configuration object containing parameters of how the model should be quantized, including mixed precision parameters.
            fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default Keras info <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/core/keras/default_framework_info.py>`_
            target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the Keras model according to.
            new_experimental_exporter (bool): Whether exporting the quantized model using new exporter or not (in progress. Avoiding it for now is recommended).

        Returns:

            A quantized model and information the user may need to handle the quantized model.

        Examples:

            Import a Keras model:

            >>> from tensorflow.keras.applications.mobilenet import MobileNet
            >>> model = MobileNet()

            Create a random dataset generator, for required number of calibration iterations (num_calibration_batches):
            In this example a random dataset of 10 batches each containing 4 images is used.

            >>> import numpy as np
            >>> num_calibration_batches = 10
            >>> def repr_datagen():
            >>>     for _ in range(num_calibration_batches):
            >>>         yield [np.random.random((4, 224, 224, 3))]

            Create an MCT core config, containing the quantization configuration:

            >>> config = mct.CoreConfig()

            If mixed precision is desired, create an MCT core config with a mixed-precision configuration, to quantize a model
            with different bitwidths for different layers.
            The candidates bitwidth for quantization should be defined in the target platform model:

            >>> config = mct.CoreConfig(mixed_precision_config=mct.MixedPrecisionQuantizationConfigV2(num_of_images=1))

            For mixed-precision set a target KPI object:
            Create a KPI object to limit our returned model's size. Note that this value affects only coefficients
            that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value,
            while the bias will not):

            >>> kpi = mct.KPI(model.count_params() * 0.75)  # About 0.75 of the model size when quantized with 8 bits.

            Create GPTQ config:

            >>> gptq_config = mct.get_keras_gptq_config(n_epochs=1)

            Pass the model with the representative dataset generator to get a quantized model:

            >>> quantized_model, quantization_info = mct.keras_gradient_post_training_quantization_experimental(model, repr_datagen, gptq_config, target_kpi=kpi, core_config=config)

        """
        KerasModelValidation(model=in_model,
                             fw_info=fw_info).validate()

        if core_config.mixed_precision_enable:
            if not isinstance(core_config.mixed_precision_config, MixedPrecisionQuantizationConfigV2):
                common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                    "MixedPrecisionQuantizationConfigV2. Please use keras_post_training_quantization API,"
                                    "or pass a valid mixed precision configuration.")

            common.Logger.info("Using experimental mixed-precision quantization. "
                               "If you encounter an issue please file a bug.")
        tb_w = _init_tensorboard_writer(fw_info)

        fw_impl = KerasImplementation()

        tg, bit_widths_config = core_runner(in_model=in_model,
                                            representative_data_gen=representative_data_gen,
                                            core_config=core_config,
                                            fw_info=fw_info,
                                            fw_impl=fw_impl,
                                            tpc=target_platform_capabilities,
                                            target_kpi=target_kpi,
                                            tb_w=tb_w)

        tg_gptq = gptq_runner(tg,
                              core_config,
                              gptq_config,
                              representative_data_gen,
                              gptq_representative_data_gen if gptq_representative_data_gen else representative_data_gen,
                              fw_info,
                              fw_impl,
                              tb_w)

        if core_config.debug_config.analyze_similarity:
            analyzer_model_quantization(representative_data_gen, tb_w, tg_gptq, fw_impl, fw_info)

        if new_experimental_exporter:
            Logger.warning('Using new experimental exported models. '
                           'Please do not use unless you are familiar with what you are doing')
            return get_exportable_keras_model(tg_gptq)

        return export_model(tg_gptq,
                            fw_info,
                            fw_impl,
                            tb_w,
                            bit_widths_config)

else:
    # If tensorflow or tensorflow_model_optimization are not installed,
    # we raise an exception when trying to use these functions.
    def get_keras_gptq_config(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_post_training_quantization_mixed_precision. '
                        'Could not find Tensorflow package.')


    def keras_gradient_post_training_quantization_experimental(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_gradient_post_training_quantization_experimental. '
                        'Could not find Tensorflow package.')



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Tuple, Callable

import tensorflow as tf
import numpy as np
from tensorflow.python.util.object_identity import Reference as TFReference

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import THRESHOLD, SIGNED, RANGE_MIN, RANGE_MAX
from model_compression_toolkit.core.common.quantization.quantizers.uniform_quantizers import threshold_is_power_of_two


def quantizer_min_max_calculator(threshold: np.ndarray,
                                 num_bits: int,
                                 signed: bool) -> Tuple[float, float]:
    """
    Compute quantization range's min/max values given a threshold, number of bits,
     and whether it's signed or not.

    Args:
        threshold: Threshold for quantization range values.
        num_bits: Number of bits to use for quantization.
        signed: Whether the quantization range should include negative values or not.

    Returns:
        Min and max values for quantization range.
    """

    if signed:
        delta = threshold / (2 ** (num_bits - 1))
        min_value = -threshold
    else:
        delta = threshold / (2 ** (num_bits))
        min_value = 0

    max_value = threshold - delta
    return min_value, max_value


def power_of_two_quantization(activation_n_bits: int,
                              quantization_params: dict) -> Callable:
    """
    Use a NodeQuantizationConfig to compute a quantizer min/max values, and use it to
    build and return a fake-quantization node with power-of-two quantization.

    Args:
        activation_n_bits: Number of bits to use for quantization.
        quantization_params: Dictionary of specific parameters for this quantization function.

    Returns:
        A fake quantization node.
    """
    activation_threshold = quantization_params.get(THRESHOLD)
    activation_is_signed = quantization_params.get(SIGNED)

    if activation_threshold is None or activation_is_signed is None:
        return None
    if not threshold_is_power_of_two(activation_threshold, per_channel=False):
        return None

    min_value, max_value = quantizer_min_max_calculator(activation_threshold,
                                                        activation_n_bits,
                                                        activation_is_signed)

    return lambda x: q(x, min_value, max_value, activation_n_bits)


def symmetric_quantization(activation_n_bits: int,
                           quantization_params: dict) -> Callable:
    """
    Use a NodeQuantizationConfig to compute a quantizer min/max values, and use it to
    build and return a symmetric fake-quantization node.

    Args:
        activation_n_bits: Number of bits to use for quantization.
        quantization_params: Dictionary of specific parameters for this quantization function.

    Returns:
        A fake quantization node.
    """
    activation_threshold = quantization_params.get(THRESHOLD)
    activation_is_signed = quantization_params.get(SIGNED)

    if activation_threshold is None or activation_is_signed is None:
        return None

    min_value, max_value = quantizer_min_max_calculator(activation_threshold,
                                                        activation_n_bits,
                                                        activation_is_signed)

    return lambda x: q(x, min_value, max_value, activation_n_bits)


def uniform_quantization(activation_n_bits: int,
                         quantization_params: dict) -> Callable:
    """
    Use a NodeQuantizationConfig to compute a quantizer min/max values, and use it to
    build and return a uniform fake-quantization node.

    Args:
        activation_n_bits: Number of bits to use for quantization.
        quantization_params: Dictionary of specific parameters for this quantization function.

    Returns:
        A fake quantization node.
    """
    min_value, max_value = quantization_params.get(RANGE_MIN), quantization_params.get(RANGE_MAX)

    if min_value is None or max_value is None:
        return None

    return lambda x: q(x, min_value, max_value, activation_n_bits)


def q(x: TFReference, min_value, max_value, activation_n_bits) -> TFReference:
    """
    Fake-quantize the input tensor x, using a tensorflow fake-quantization node.

    Args:
        x: Input tensor to quantize.
        min_value: quantization range lower bound.
        max_value: quantization range upper bound.
        activation_n_bits: Number of bits to use for quantization.

    Returns:
        The fake-quantized input tensor.
    """
    if x.dtype != tf.float32:
        x = tf.cast(x, dtype=tf.float32)

    # fake_quant_with_min_max_vars expects to get x of float32
    return tf.quantization.fake_quant_with_min_max_vars(x,
                                                        min=min_value,
                                                        max=max_value,
                                                        num_bits=activation_n_bits)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from model_compression_toolkit.core.common.constants import BITS_TO_BYTES


def compute_quantize_tensor_memory_bytes(tensor_size: float, n_bits: int) -> float:
    """
    A utility function to compute the actual memory size of a tensor for a given bit-width.

    Args:
        tensor_size: The number of parameters in the tensor.
        n_bits: The bit-width in which the tensor values are represented.

    Returns: The size of the tensor in memory in bytes.

    """
    return tensor_size * n_bits / BITS_TO_BYTES



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Any
from model_compression_toolkit.core.common.matchers.node_matcher import BaseNodeMatcher


class NodeTypeFilter(BaseNodeMatcher):
    """
    Class NodeNameFilter to check if a node is of a specific type.
    """
    def __init__(self, node_type):
        """
        Init a NodeTypeFilter object.

        Args:
            node_type: Node type to check.
        """
        self.node_type = node_type

    def apply(self, input_object: Any) -> bool:
        """
        Check if input_object is of the type that NodeTypeFilter contains.

        Args:
            input_object: Node object to check for its type.

        Returns:
            True if the node if of the type that was passed during the initialization of NodeTypeFilter.
        """
        if input_object.type == self.node_type:
            return True


class NodeNameFilter(BaseNodeMatcher):
    """
    Class NodeNameFilter to check if a node's name has a specific value.
    """
    def __init__(self, node_name):
        """
        Init a NodeNameFilter object.

        Args:
            node_name: Node name to check.
        """
        self.node_name = node_name

    def apply(self, input_object: Any) -> bool:
        """
        Check if input_object's node name is the name that NodeNameFilter contains.

        Args:
            input_object: Node object to check for its name.

        Returns:
            True if the node's name is tha same as the name that was passed during the initialization of NodeNameFilter.
        """
        if input_object.name == self.node_name:
            return True


class NodeNameScopeFilter(BaseNodeMatcher):
    """
    Class NodeNameFilter to check if a string is in a node's name.
    """
    def __init__(self, node_name_scope):
        """
        Init a NodeNameScopeFilter object.

        Args:
            node_name_scope: String to check if exists in node name.
        """
        self.node_name_scope = node_name_scope

    def apply(self, input_object: Any) -> bool:
        """
        Check if input_object's node name contains the string NodeNameScopeFilter contains.

        Args:
            input_object: Node object to check its name contains the string.

        Returns:
            True if the node's name contains the string that was passed during the initialization of NodeNameScopeFilter.
        """
        if self.node_name_scope in input_object.name:
            return True



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


def get_current_tpc():
    """

    Returns: The current TargetPlatformCapabilities that is being used and accessed.

    """
    return _current_tpc.get()


class _CurrentTPC(object):
    """
    Wrapper of the current TargetPlatformCapabilities object that is being accessed and defined.
    """
    def __init__(self):
        super(_CurrentTPC, self).__init__()
        self.tpc = None

    def get(self):
        """

        Returns: The current TargetPlatformCapabilities that is being defined.

        """
        if self.tpc is None:
            raise Exception('TargetPlatformCapabilities is not initialized.')
        return self.tpc

    def reset(self):
        """

        Reset the current TargetPlatformCapabilities so a new TargetPlatformCapabilities can be wrapped and
        used as the current TargetPlatformCapabilities object.

        """
        self.tpc = None

    def set(self, target_platform_capabilities):
        """
        Set and wrap a TargetPlatformCapabilities as the current TargetPlatformCapabilities.

        Args:
            target_platform_capabilities: TargetPlatformCapabilities to set as the current TargetPlatformCapabilities
            to access and use.

        """
        self.tpc = target_platform_capabilities


# Use a single instance for the current model.
_current_tpc = _CurrentTPC()



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import numpy as np
import torch
import torch.nn as nn
import operator
from typing import List

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.graph.base_graph import Graph, BaseNode, OutTensor
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.pytorch.constants import KERNEL, BIAS, NUM_HEADS, KEY_DIM, VALUE_DIM, \
    EMBED_DIM, BIAS_K, BIAS_V, ADD_ZERO_ATTN, BATCH_FIRST, OUT_PROJ_WEIGHT, OUT_PROJ_BIAS, \
    V_PROJ_WEIGHT, K_PROJ_WEIGHT, Q_PROJ_WEIGHT, IN_PROJ_WEIGHT, IN_PROJ_BIAS, DIM, KERNEL_SIZE, \
    IN_CHANNELS, OUT_CHANNELS


class MHAParams:
    """
    A data class to hold all relevant parameters from the MHA node framework attributes
    """

    def __init__(self,
                 mha_node: BaseNode):
        """
        Extract MHA params from layer attributes
        Args:
            mha_node: MHA node
        """

        # Only batch first network is supported
        if BATCH_FIRST in mha_node.framework_attr.keys():
            if mha_node.framework_attr[BATCH_FIRST] is not True:
                raise Exception('Only batch first network is supported')
        else:
            raise Exception('Only batch first network is supported')

        # Add Zero Attn feature is Not Implemented
        if ADD_ZERO_ATTN in mha_node.framework_attr.keys():
            if mha_node.framework_attr[ADD_ZERO_ATTN] is not False:
                raise Exception('Add Zero Attn feature is Not Implemented')

        # Check if Add Bias KV feature is Active
        if BIAS_K and BIAS_V in mha_node.weights.keys():
            if mha_node.weights[BIAS_K] and mha_node.weights[BIAS_V] is not None:
                raise Exception('Add BIAS_KV feature is Not Implemented')

        self.embed_dim = mha_node.framework_attr[EMBED_DIM]
        self.num_heads = mha_node.framework_attr[NUM_HEADS]

        if KEY_DIM in mha_node.framework_attr:
            self.kdim = mha_node.framework_attr[KEY_DIM]
        else:
            self.kdim = False

        if VALUE_DIM in mha_node.framework_attr:
            self.vdim = mha_node.framework_attr[VALUE_DIM]
        else:
            self.vdim = False

        self.qdim = int(self.embed_dim / self.num_heads)

        self.q_input, self.k_input, self.v_input = mha_node.input_shape

        # check for input correctness
        assert self.q_input[0] == self.k_input[0] == self.v_input[0], "Batch size must be equal to all inputs"
        assert self.k_input[1] == self.v_input[1], "key's sequence dim do not match value's"
        assert self.kdim == self.k_input[2], "key's embed dim do not match excepted"
        assert self.vdim == self.v_input[2], "value's embed dim do not match excepted"

        self.kv_seq = self.k_input[1]
        self.q_seq = self.q_input[1]

        self.q_input_shape = tuple(self.q_input)
        self.q_transpose_in_shape = tuple([self.q_input[0], self.embed_dim, self.q_seq])
        self.q_reshape_in_shape = tuple([self.q_input[0], self.num_heads, self.qdim, self.q_seq])
        self.q_transpose_shape = tuple([self.q_input[0], self.num_heads, self.q_seq, self.qdim])
        self.q_split_shape = tuple([self.q_input[0], 1, self.q_seq, self.qdim])
        self.q_head_shape = tuple([self.q_input[0], self.q_seq, self.qdim])

        self.k_input_shape = tuple(self.k_input)
        self.k_transpose_in_shape = tuple([self.k_input[0], self.kdim, self.kv_seq])
        self.k_proj_in_shape = tuple([self.k_input[0], self.embed_dim, self.kv_seq])
        self.k_reshape_in_shape = tuple([self.k_input[0], self.num_heads, self.qdim, self.kv_seq])
        self.k_split_shape = tuple([self.k_input[0], 1, self.qdim, self.kv_seq])
        self.k_head_shape = tuple([self.k_input[0], self.qdim, self.kv_seq])

        self.v_input_shape = tuple(self.v_input)
        self.v_transpose_in_shape = tuple([self.v_input[0], self.vdim, self.kv_seq])
        self.v_proj_in_shape = tuple([self.v_input[0], self.embed_dim, self.kv_seq])
        self.v_reshape_in_shape = tuple([self.v_input[0], self.num_heads, self.qdim, self.kv_seq])
        self.v_transpose_shape = tuple([self.v_input[0], self.num_heads, self.kv_seq, self.qdim])
        self.v_split_shape = tuple([self.v_input[0], 1, self.kv_seq, self.qdim])
        self.v_head_shape = tuple([self.v_input[0], self.kv_seq, self.qdim])

        self.attn_mat_shape = tuple([self.q_input[0], 1, self.q_seq, self.kv_seq])
        self.attn_shape = self.q_split_shape

        self.attn_cat_shape = tuple([self.q_input[0], self.num_heads, self.q_seq, self.qdim])
        self.attn_transpose_shape = tuple([self.q_input[0], self.q_seq, self.num_heads, self.qdim])
        self.attn_reshape_shape = self.q_input_shape

        self.transpose_out_proj_shape = self.q_transpose_in_shape
        self.output_shape = self.q_input_shape


class MultiHeadAttentionDecomposition(common.BaseSubstitution):
    """
    Removes a MultiHeadAttention node from the graph,
    and replaces it with a compatible graph that consists of Conv, MatMul, Softmax and stacked layers.
    """

    def __init__(self):
        """
        Matches MultiHeadAttention node.
        """
        super().__init__(matcher_instance=NodeOperationMatcher(nn.MultiheadAttention))

    def _project_input(self,
                       graph: Graph,
                       mha_node: BaseNode,
                       params: MHAParams) -> List[BaseNode]:
        """
        This method creates the nodes required to project q, k, v
        We implement the projection as Conv1d
        Due to the above we add transpose node to each of the inputs in order to swap the channel axis according
        to Conv1d expected input shape
        We describe below the shape transformation of each input (q, k, v) from the input shape,
        to transposing and projecting

        Args:
            graph: Graph to apply the substitution on.
            mha_node: MHA node.
            params: MHAnode params.

        Returns:
            List of nodes transposing the inputs and nodes after input projection.
        """
        # q shape transformation:
        # (B, q_seq, q_dim*n_h) --> (B, q_dim*n_h, q_seq) --> (B, q_dim*n_h, q_seq)
        #
        # k shape transformation:
        # (B, kv_seq, k_dim) --> (B, k_dim, kv_seq) --> (B, q_dim*n_h, kv_seq)
        #
        # v shape transformation:
        # (B, kv_seq, v_dim) --> (B, v_dim, kv_seq) --> (B, q_dim*n_h, kv_seq)

        # The scaling const in MHA, we implement the scaling in q input projection weight
        factor = (params.qdim ** -0.5)

        # In some cases (where k_dim and v_dim params isn't supplied) the input projection weight
        # located under 'in_proj_weight' to all inputs (q, k, v) and needs to be separated by 3 inputs
        if IN_PROJ_WEIGHT in mha_node.weights.keys():
            in_proj_w = np.expand_dims(mha_node.get_weights_by_keys(IN_PROJ_WEIGHT).copy(), -1)
            qk, kk, vk = np.split(in_proj_w, 3)
            qk = qk * factor
        # In other cases the weight of the input projection located under 'q_proj_weight' for q input projection and
        # 'k_proj_weight' for k input projection etc
        else:
            qk = np.expand_dims(mha_node.get_weights_by_keys(Q_PROJ_WEIGHT).copy() * factor, -1)
            kk = np.expand_dims(mha_node.get_weights_by_keys(K_PROJ_WEIGHT).copy(), -1)
            vk = np.expand_dims(mha_node.get_weights_by_keys(V_PROJ_WEIGHT).copy(), -1)
        # If the input projection has bias, it'll be located under  'in_proj_bias' for all inputs (q, k, v)
        # and needs to be separated by 3 inputs
        if IN_PROJ_BIAS in mha_node.weights.keys():
            in_proj_bias = mha_node.get_weights_by_keys(IN_PROJ_BIAS).copy()
            qb, kb, vb = np.split(in_proj_bias, 3)
            qb = qb * factor
            in_bias = True
            q_node_weights = {KERNEL: qk, BIAS: qb}
            k_node_weights = {KERNEL: kk, BIAS: kb}
            v_node_weights = {KERNEL: vk, BIAS: vb}
        else:
            in_bias = False
            q_node_weights = {KERNEL: qk}
            k_node_weights = {KERNEL: kk}
            v_node_weights = {KERNEL: vk}

        # project query, key, value

        query_name = f'{mha_node.name}_query_in_proj'
        # Due to the input projection implementation as Conv1d we need to swap the channel axis according from
        # last axis to the second
        # Transpose q second and last axis
        # (B, q_seq, q_dim*n_h) --> (B, q_dim*n_h, q_seq)
        q_transpose_node = FunctionalNode(name=query_name + '_transpose',
                                          framework_attr={},
                                          input_shape=params.q_input_shape,
                                          output_shape=params.q_transpose_in_shape,
                                          weights={},
                                          layer_class=torch.transpose,
                                          op_call_args=[1, 2],
                                          op_call_kwargs={},
                                          functional_op=torch.transpose)
        graph.add_node(q_transpose_node)

        # q doesn't change his shape during projection
        # (B, q_dim*n_h, q_seq)  --> (B, q_dim*n_h, q_seq)
        q_node = BaseNode(name=query_name,
                          framework_attr={IN_CHANNELS: params.embed_dim, OUT_CHANNELS: params.embed_dim,
                                          KERNEL_SIZE: 1, BIAS: in_bias},
                          input_shape=params.q_transpose_in_shape,
                          output_shape=params.q_transpose_in_shape,
                          weights=q_node_weights,
                          layer_class=nn.Conv1d)
        graph.add_node_with_in_edges(q_node, [q_transpose_node])

        key_name = f'{mha_node.name}_key_in_proj'
        # Due to the input projection implementation as Conv1d we need to swap the channel axis according from
        # last axis to the second
        # Transpose k second and last axis
        # (B, k_seq, k_dim) --> (B, k_dim, k_seq)
        k_transpose_node = FunctionalNode(name=key_name + '_transpose',
                                          framework_attr={},
                                          input_shape=params.k_input_shape,
                                          output_shape=params.k_transpose_in_shape,
                                          weights={},
                                          layer_class=torch.transpose,
                                          op_call_args=[1, 2],
                                          op_call_kwargs={},
                                          functional_op=torch.transpose)
        graph.add_node(k_transpose_node)

        # k shape is changing due to projection
        # (B, k_dim, k_seq) --> (B, q_dim*n_h, k_seq)
        k_node = BaseNode(name=key_name,
                          framework_attr={IN_CHANNELS: params.kdim, OUT_CHANNELS: params.embed_dim,
                                          KERNEL_SIZE: 1, BIAS: in_bias},
                          input_shape=params.k_transpose_in_shape,
                          output_shape=params.k_proj_in_shape,
                          weights=k_node_weights,
                          layer_class=nn.Conv1d)
        graph.add_node_with_in_edges(k_node, [k_transpose_node])

        value_name = f'{mha_node.name}_value_in_proj'
        # Due to the input projection implementation as Conv1d we need to swap the channel axis according from
        # last axis to the second
        # Transpose v second and last axis
        # (B, v_seq, v_dim) --> (B, v_dim, v_seq)
        v_transpose_node = FunctionalNode(name=value_name + '_transpose',
                                          framework_attr={},
                                          input_shape=params.v_input_shape,
                                          output_shape=params.v_transpose_in_shape,
                                          weights={},
                                          layer_class=torch.transpose,
                                          op_call_args=[1, 2],
                                          op_call_kwargs={},
                                          functional_op=torch.transpose)
        graph.add_node(v_transpose_node)

        # v shape is changing due to projection
        # (B, v_seq, v_dim) --> (B, q_dim*n_h, v_seq)
        v_node = BaseNode(name=value_name,
                          framework_attr={IN_CHANNELS: params.vdim, OUT_CHANNELS: params.embed_dim,
                                          KERNEL_SIZE: 1, BIAS: in_bias},
                          input_shape=params.v_transpose_in_shape,
                          output_shape=params.v_proj_in_shape,
                          weights=v_node_weights,
                          layer_class=nn.Conv1d)
        graph.add_node_with_in_edges(v_node, [v_transpose_node])

        return q_transpose_node, k_transpose_node, v_transpose_node, q_node, k_node, v_node

    @staticmethod
    def _arrange_before_split(graph: Graph,
                              mha_node: BaseNode,
                              q_node: BaseNode,
                              k_node: BaseNode,
                              v_node: BaseNode,
                              params: MHAParams) -> List[BaseNode]:
        """
        This method creates the nodes required for arranging the shapes of q, k, v, after
        the input projection, before the split by head operation.

        Args:
            graph: Graph to apply the substitution on.
            mha_node: MHA node.
            q_node: query node after input projection.
            k_node: key node after input projection.
            v_node: value node after input projection.
            params: MHAnode params.

        Returns:
            List of nodes after shape arranging.
        """
        # (B, q_dim*n_h, q_seq)  --> (B, n_h, q_dim, q_seq)  --> (B, n_h, q_seq, q_dim)
        # (B, q_dim*n_h, kv_seq) -->  (B, n_h, q_dim, kv_seq)
        # (B, q_dim*n_h, kv_seq) -->  (B, n_h, q_dim, kv_seq) --> (B, n_h, kv_seq, q_dim)

        query_name = f'{mha_node.name}_query'
        # (B, q_dim*n_h, q_seq) --> (B, n_h, q_dim, q_seq)
        q_reshape_node = FunctionalNode(name=query_name + '_reshape',
                                        framework_attr={},
                                        input_shape=params.q_transpose_in_shape,
                                        output_shape=params.q_reshape_in_shape,
                                        weights={},
                                        layer_class=torch.reshape,
                                        op_call_args=[params.q_reshape_in_shape],
                                        op_call_kwargs={},
                                        functional_op=torch.reshape)
        graph.add_node_with_in_edges(q_reshape_node, [q_node])

        # (B, n_h, q_dim, q_seq) --> (B, n_h, q_seq, q_dim)
        q_transpose_node = FunctionalNode(name=query_name + '_transpose',
                                          framework_attr={},
                                          input_shape=params.q_reshape_in_shape,
                                          output_shape=params.q_transpose_shape,
                                          weights={},
                                          layer_class=torch.transpose,
                                          op_call_args=[2, 3],
                                          op_call_kwargs={},
                                          functional_op=torch.transpose)
        graph.add_node_with_in_edges(q_transpose_node, [q_reshape_node])

        key_name = f'{mha_node.name}_key'
        # (B, q_dim*n_h, kv_seq) -->  (B, n_h, q_dim, kv_seq)
        k_reshape_node = FunctionalNode(name=key_name + '_reshape',
                                        framework_attr={},
                                        input_shape=params.k_proj_in_shape,
                                        output_shape=params.k_reshape_in_shape,
                                        weights={},
                                        layer_class=torch.reshape,
                                        op_call_args=[params.k_reshape_in_shape],
                                        op_call_kwargs={},
                                        functional_op=torch.reshape)
        graph.add_node_with_in_edges(k_reshape_node, [k_node])

        value_name = f'{mha_node.name}_value'
        # (B, q_dim*n_h, kv_seq) -->  (B, n_h, q_dim, kv_seq)
        v_reshape_node = FunctionalNode(name=value_name + '_reshape',
                                        framework_attr={},
                                        input_shape=params.v_proj_in_shape,
                                        output_shape=params.v_reshape_in_shape,
                                        weights={},
                                        layer_class=torch.reshape,
                                        op_call_args=[params.v_reshape_in_shape],
                                        op_call_kwargs={},
                                        functional_op=torch.reshape)
        graph.add_node_with_in_edges(v_reshape_node, [v_node])

        # (B, n_h, q_dim, kv_seq) --> (B, n_h, kv_seq, q_dim)
        v_transpose_node = FunctionalNode(name=value_name + '_transpose',
                                          framework_attr={},
                                          input_shape=params.v_reshape_in_shape,
                                          output_shape=params.v_transpose_shape,
                                          weights={},
                                          layer_class=torch.transpose,
                                          op_call_args=[2, 3],
                                          op_call_kwargs={},
                                          functional_op=torch.transpose)
        graph.add_node_with_in_edges(v_transpose_node, [v_reshape_node])

        return q_transpose_node, k_reshape_node, v_transpose_node

    @staticmethod
    def _split_projected(graph: Graph,
                         name: str,
                         q_transpose_node: BaseNode,
                         k_reshape_node: BaseNode,
                         v_transpose_node: BaseNode,
                         params: MHAParams) -> List[BaseNode]:
        """
        This method creates the nodes required for splitting q, k, v to query, key and value per head
        (total of num_heads q, k and v).

        Args:
            graph: Graph to apply the substitution on.
            name: MHA node name.
            q_transpose_node: query node after shape arranging.
            k_reshape_node: key node after shape arranging.
            v_transpose_node: value node after shape arranging.
            params: MHAnode params.

        Returns:
            List of nodes after splitting.
        """
        # Split q, k, v to query, key and value per head (total of num_heads q, k and v)
        # (B, n_h, q_seq, q_dim) --> (B, 1, q_seq, q_dim)* n_h
        # (B, n_h, q_dim, kv_seq) --> (B, 1, q_dim, kv_seq) * n_h
        # (B, n_h, kv_seq, q_dim) --> (B, 1, kv_seq, q_dim) * n_h

        query_name = f'{name}_query'
        # (B, n_h, q_seq, q_dim) --> (B, 1, q_seq, q_dim) * n_h
        q_split_node = FunctionalNode(name=query_name + '_split',
                                      framework_attr={DIM: 1},
                                      input_shape=params.q_transpose_shape,
                                      output_shape=[params.q_split_shape] * params.num_heads,
                                      weights={},
                                      layer_class=torch.split,
                                      op_call_args=[1],
                                      op_call_kwargs={DIM: 1},
                                      functional_op=torch.split)
        graph.add_node_with_in_edges(q_split_node, [q_transpose_node])

        key_name = f'{name}_key'
        # (B, n_h, q_dim, kv_seq) --> (B, 1, q_dim, kv_seq) * n_h
        k_split_node = FunctionalNode(name=key_name + '_split',
                                      framework_attr={DIM: 1},
                                      input_shape=params.k_reshape_in_shape,
                                      output_shape=[params.k_split_shape] * params.num_heads,
                                      weights={},
                                      layer_class=torch.split,
                                      op_call_args=[1],
                                      op_call_kwargs={DIM: 1},
                                      functional_op=torch.split)
        graph.add_node_with_in_edges(k_split_node, [k_reshape_node])

        value_name = f'{name}_value'
        # (B, n_h, kv_seq, q_dim) --> (B, 1, kv_seq, q_dim) * n_h
        v_split_node = FunctionalNode(name=value_name + '_split',
                                      framework_attr={DIM: 1},
                                      input_shape=params.v_reshape_in_shape,
                                      output_shape=[params.v_split_shape] * params.num_heads,
                                      weights={},
                                      layer_class=torch.split,
                                      op_call_args=[1],
                                      op_call_kwargs={DIM: 1},
                                      functional_op=torch.split)
        graph.add_node_with_in_edges(v_split_node, [v_transpose_node])

        return q_split_node, k_split_node, v_split_node

    @staticmethod
    def _calc_attention_head(graph: Graph,
                             q_in_node: BaseNode,
                             k_in_node: BaseNode,
                             v_in_node: BaseNode,
                             mha_node: BaseNode,
                             head_index: int,
                             params: MHAParams) -> BaseNode:
        """
        This method creates the nodes required for attention calc by head

        Args:
            graph: Graph to apply the substitution on.
            q_in_node: query node after shape arranging.
            k_in_node: key node after shape arranging.
            v_in_node: value node after shape arranging.
            mha_node: MHA node.
            head_index: index of the head.
            params: MHAnode params.

        Returns:
            Node after attention calc.
        """
        # Q X K = attn
        # (B, 1, q_seq, q_dim) X (B, 1, q_dim, kv_seq) = (B, 1, q_seq, kv_seq)

        # attn X V = attn_out
        # (B, 1, q_seq, kv_seq) X (B, 1, kv_seq, q_dim) = (B, 1, q_seq, q_dim)

        # (B, 1, q_seq, q_dim) * n_h --> (B, 1, q_seq, kv_seq)
        get_q_name = f'{mha_node.name}_get_q_{head_index}'
        get_q_node = FunctionalNode(name=get_q_name,
                                    framework_attr={},
                                    input_shape=(params.q_split_shape,) * params.num_heads,
                                    output_shape=params.q_split_shape,
                                    weights={},
                                    layer_class=operator.getitem,
                                    op_call_args=[head_index],
                                    op_call_kwargs={},
                                    functional_op=operator.getitem)
        graph.add_node_with_in_edges(get_q_node, [q_in_node], [head_index])

        # (B, 1, q_seq, q_dim) * n_h --> (B, 1, q_seq, kv_seq)
        get_k_name = f'{mha_node.name}_get_k_{head_index}'
        get_k_node = FunctionalNode(name=get_k_name,
                                    framework_attr={},
                                    input_shape=(params.k_split_shape,) * params.num_heads,
                                    output_shape=params.k_split_shape,
                                    weights={},
                                    layer_class=operator.getitem,
                                    op_call_args=[head_index],
                                    op_call_kwargs={},
                                    functional_op=operator.getitem)
        graph.add_node_with_in_edges(get_k_node, [k_in_node], [head_index])

        # (B, 1, q_seq, q_dim) X (B, 1, q_dim, kv_seq) = (B, 1, q_seq, kv_seq)
        matmul_name = f'{mha_node.name}_matmul_{head_index}'
        matmul_node = FunctionalNode(name=matmul_name,
                                     framework_attr={},
                                     input_shape=(params.q_split_shape, params.k_split_shape),
                                     output_shape=params.attn_mat_shape,
                                     weights={},
                                     layer_class=torch.matmul,
                                     op_call_args=[],
                                     op_call_kwargs={},
                                     functional_op=torch.matmul)
        graph.add_node_with_in_edges(matmul_node, [get_q_node, get_k_node])

        # apply softmax on attention matrix
        softmax_name = f'{mha_node.name}_softmax_{head_index}'
        softmax_node = BaseNode(name=softmax_name,
                                framework_attr={DIM: -1},
                                input_shape=params.attn_mat_shape,
                                output_shape=params.attn_mat_shape,
                                weights={},
                                layer_class=nn.Softmax)
        graph.add_node_with_in_edges(softmax_node, [matmul_node])

        # (B, 1, q_seq, q_dim) * n_h --> (B, 1, q_seq, kv_seq)
        get_v_name = f'{mha_node.name}_get_v_{head_index}'
        get_v_node = FunctionalNode(name=get_v_name,
                                    framework_attr={},
                                    input_shape=(params.v_split_shape,) * params.num_heads,
                                    output_shape=params.v_split_shape,
                                    weights={},
                                    layer_class=operator.getitem,
                                    op_call_args=[head_index],
                                    op_call_kwargs={},
                                    functional_op=operator.getitem)
        graph.add_node_with_in_edges(get_v_node, [v_in_node], [head_index])

        # (B, 1, q_seq, kv_seq) X (B, 1, kv_seq, q_dim) = (B, 1, q_seq, q_dim)
        matmulv_name = f'{mha_node.name}_dotv_{head_index}'
        matmulv_node = FunctionalNode(name=matmulv_name,
                                      framework_attr={},
                                      input_shape=(params.attn_mat_shape, params.v_split_shape),
                                      output_shape=params.attn_shape,
                                      weights={},
                                      layer_class=torch.matmul,
                                      op_call_args=[],
                                      op_call_kwargs={},
                                      functional_op=torch.matmul)
        graph.add_node_with_in_edges(matmulv_node, [softmax_node, get_v_node])

        return matmulv_node

    @staticmethod
    def _cat_heads_reshape(graph: Graph,
                           name: str,
                           att_head_output_nodes: List[BaseNode],
                           params: MHAParams) -> BaseNode:
        """
        This method creates the nodes required for concatenating all heads after attention

        Args:
            graph: Graph to apply the substitution on.
            name: MHA node name.
            att_head_output_nodes: list of nodes after attention.
            params: MHAnode params.

        Returns:
            Node after cat and reshape.
        """
        # [(B, 1, q_seq, q_dim)]  * n_h --> (B, n_h, q_seq, q_dim)
        # -->  (B, q_seq, n_h, q_dim)  -->  (B, q_seq, q_dim*n_h)

        # [(B, 1, q_seq, q_dim)]  * n_h --> (B, n_h, q_seq, q_dim)
        cat_node = FunctionalNode(name=f'{name}_cat',
                                  framework_attr={DIM: 1},
                                  input_shape=(params.attn_shape,) * params.num_heads,
                                  output_shape=params.attn_cat_shape,
                                  weights={},
                                  layer_class=torch.cat,
                                  op_call_args=[],
                                  op_call_kwargs={DIM: 1},
                                  functional_op=torch.cat, inputs_as_list=True)
        graph.add_node_with_in_edges(cat_node, att_head_output_nodes)

        # (B, n_h, q_seq, q_dim) -->  (B, q_seq, n_h, q_dim)
        transpose_node = FunctionalNode(name=f'{name}_transpose',
                                        framework_attr={},
                                        input_shape=params.attn_cat_shape,
                                        output_shape=params.attn_transpose_shape,
                                        weights={},
                                        layer_class=torch.transpose,
                                        op_call_args=[1, 2],
                                        op_call_kwargs={},
                                        functional_op=torch.transpose)
        graph.add_node_with_in_edges(transpose_node, [cat_node])

        # (B, q_seq, n_h, q_dim)  -->  (B, q_seq, q_dim*n_h)
        attn_reshape_node = FunctionalNode(name=f'{name}_attn_reshape',
                                           framework_attr={},
                                           input_shape=params.attn_transpose_shape,
                                           output_shape=params.attn_reshape_shape,
                                           weights={},
                                           layer_class=torch.reshape,
                                           op_call_args=[params.attn_reshape_shape],
                                           op_call_kwargs={},
                                           functional_op=torch.reshape)
        graph.add_node_with_in_edges(attn_reshape_node, [transpose_node])

        return attn_reshape_node

    def _project_output(self,
                        graph: Graph,
                        mha_node: BaseNode,
                        attn_reshape_node: BaseNode,
                        params: MHAParams) -> BaseNode:
        """
        This method creates the nodes required for output projecting

        Args:
            graph: Graph to apply the substitution on.
            mha_node: MHA node.
            attn_reshape_node: attention node.
            params: MHAnode params.

        Returns:
            Node after projection.
        """

        # (B, q_seq, q_dim*n_h) -->  (B, q_dim*n_h, q_seq) -->  (B, q_dim*n_h, q_seq) --> (B, q_seq, q_dim*n_h)

        outk = np.expand_dims(mha_node.get_weights_by_keys(OUT_PROJ_WEIGHT).copy(), -1)
        if OUT_PROJ_BIAS in mha_node.weights.keys():
            outb = mha_node.get_weights_by_keys(OUT_PROJ_BIAS).copy()
            out_bias = True
            out_proj_node_weights = {KERNEL: outk, BIAS: outb}
        else:
            out_bias = False
            out_proj_node_weights = {KERNEL: outk}

        # project out
        out_name = f'{mha_node.name}_project_out'

        # transpose proj out
        # (B, q_seq, q_dim*n_h) -->  (B, q_dim*n_h, q_seq)
        transpose_node = FunctionalNode(name=out_name + '_transpose',
                                        framework_attr={},
                                        input_shape=params.attn_reshape_shape,
                                        output_shape=params.transpose_out_proj_shape,
                                        weights={},
                                        layer_class=torch.transpose,
                                        op_call_args=[1, 2],
                                        op_call_kwargs={},
                                        functional_op=torch.transpose)
        graph.add_node_with_in_edges(transpose_node, [attn_reshape_node])

        # (B, q_dim*n_h, q_seq) -->  (B, q_dim*n_h, q_seq)
        proj_out_node = BaseNode(name=out_name,
                                 framework_attr={IN_CHANNELS: params.embed_dim, OUT_CHANNELS: params.embed_dim,
                                                 KERNEL_SIZE: 1, BIAS: out_bias},
                                 input_shape=params.transpose_out_proj_shape,
                                 output_shape=params.transpose_out_proj_shape,
                                 weights=out_proj_node_weights,
                                 layer_class=nn.Conv1d)
        graph.add_node_with_in_edges(proj_out_node, [transpose_node])

        # transpose output
        # (B, q_dim*n_h, q_seq) -->  (B, q_seq, q_dim*n_h)
        transpose_out_node = FunctionalNode(name=f'{mha_node.name}_transpose_output',
                                            framework_attr={},
                                            input_shape=params.transpose_out_proj_shape,
                                            output_shape=params.output_shape,
                                            weights={},
                                            layer_class=torch.transpose,
                                            op_call_args=[1, 2],
                                            op_call_kwargs={},
                                            functional_op=torch.transpose)
        graph.add_node_with_in_edges(transpose_out_node, [proj_out_node])
        return transpose_out_node

    @staticmethod
    def _connect_to_graph(graph: Graph,
                          mha_node: BaseNode,
                          q_node: BaseNode,
                          k_node: BaseNode,
                          v_node: BaseNode,
                          output_permute_node: BaseNode):
        """
        connect subgraph to input graph
        Args:
            graph: input graph
            mha_node: MHA node to substitute inputs and outputs with
            q_node: 1st input to MHA node
            k_node: 2nd input to MHA node
            v_node: 3rd input to MHA node
            output_permute_node: output node of MHA node
        """
        query_in_edge, key_in_edge, value_in_edge = graph.in_edges(mha_node)
        graph.add_edge(query_in_edge[0], q_node, **graph.get_edge_data(*query_in_edge, 0))
        graph.add_edge(key_in_edge[0], k_node, **graph.get_edge_data(*key_in_edge, 0))
        graph.add_edge(value_in_edge[0], v_node, **graph.get_edge_data(*value_in_edge, 0))
        graph.remove_edge(query_in_edge[0], mha_node)
        graph.remove_edge(key_in_edge[0], mha_node)
        graph.remove_edge(value_in_edge[0], mha_node)
        graph.reconnect_out_edges(current_node=mha_node, new_node=output_permute_node)

    def substitute(self,
                   graph: Graph,
                   mha_node: BaseNode) -> Graph:
        """
        connect subgraph to input graph
        Args:
            graph: input graph
            mha_node: MHA node to substitute inputs and outputs with
        Returns:
            Graph after applying the substitution.
        """

        if mha_node.reuse:
            raise Exception("MCT doesn't support reuse of MultiHeadAttention layer")
        params = MHAParams(mha_node)

        # project
        # (B, q_seq, q_dim*n_h) --> (B, q_dim*n_h, q_seq)
        # (B, kv_seq, k_dim) --> (B, q_dim*n_h, kv_seq)
        # (B, kv_seq, v_dim) --> (B, q_dim*n_h, kv_seq)
        q_transpose_node, k_transpose_node, v_transpose_node, \
        q_node, k_node, v_node = self._project_input(graph, mha_node, params)

        # Arrange k, q, v before slicing per head
        # (B, q_dim*n_h, q_seq)  --> (B, n_h, q_seq, q_dim)
        # (B, q_dim*n_h, kv_seq) --> (B, n_h, q_dim, kv_seq)
        # (B, q_dim*n_h, kv_seq) --> (B, n_h, kv_seq, q_dim)
        q_fixed_node, k_fixed_node, v_fixed_node = self._arrange_before_split(graph, mha_node, q_node, k_node, v_node,
                                                                              params)
        # (B, n_h, q_seq, q_dim) --> (B, 1, q_seq, q_dim) * 3
        # (B, n_h, q_dim, kv_seq) --> (B, 1, q_dim, kv_seq) * 3
        # (B, n_h, kv_seq, q_dim) --> (B, 1, kv_seq, q_dim) * 3
        q_split_node, k_split_node, v_split_node = self._split_projected(graph, mha_node.name, q_fixed_node,
                                                                         k_fixed_node, v_fixed_node, params)

        att_head_output_nodes = []
        for h in range(params.num_heads):
            # (B, 1, q_seq, q_dim) X (B, 1, q_dim, kv_seq) = (B, 1, q_seq, kv_seq)
            # Apply Softmax
            # (B, 1, q_seq, kv_seq) X (B, 1, kv_seq, q_dim) = (B, 1, q_seq, q_dim)
            dotv_node = self._calc_attention_head(graph, q_split_node, k_split_node, v_split_node,
                                                  mha_node, h, params)
            # [(B, 1, q_seq, q_dim)] * n_h
            att_head_output_nodes.append(dotv_node)

        # [(B, 1, q_seq, q_dim)]  * n_h -->  (B, q_seq, q_dim*n_h)
        attn_reshape_node = self._cat_heads_reshape(graph, mha_node.name, att_head_output_nodes, params)

        # (B, q_seq, q_dim*n_h) --> (B, q_seq, q_dim*n_h)
        proj_output = self._project_output(graph, mha_node, attn_reshape_node, params)

        # connect edges to new nodes
        self._connect_to_graph(graph, mha_node, q_transpose_node, k_transpose_node, v_transpose_node, proj_output)

        # Finally, remove the MHA node
        graph.remove_node(mha_node, new_graph_outputs=[OutTensor(proj_output, 0)])

        return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict, Callable

import keras.models
import keras.models
import tensorflow as tf
from keras.engine.base_layer import Layer
from tensorflow_model_optimization.python.core.quantization.keras.default_8bit.default_8bit_quantize_configs import \
    NoOpQuantizeConfig

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.exporter.model_exporter.keras.base_keras_exporter import \
    BaseKerasExporter
from model_compression_toolkit.exporter.model_wrapper.keras.builder.quantize_config_to_node import \
    SUPPORTED_QUANTIZATION_CONFIG
from model_compression_toolkit.exporter.model_wrapper.keras.extended_quantize_wrapper import ExtendedQuantizeWrapper
from model_compression_toolkit.exporter.model_wrapper.keras.quantize_configs.activation_quantize_config import \
    ActivationQuantizeConfig
from model_compression_toolkit.exporter.model_wrapper.keras.quantize_configs.weights_activation_quantize_config \
    import \
    WeightsActivationQuantizeConfig
from model_compression_toolkit.exporter.model_wrapper.keras.quantize_configs.weights_quantize_config import \
    WeightsQuantizeConfig
from model_compression_toolkit.exporter.model_wrapper.keras.quantizers.fq_quantizer import FakeQuantQuantizer


class FakelyQuantKerasExporter(BaseKerasExporter):
    """
    Exporter for fakely-quant Keras models.
    The exporter expects to receive an exportable model (where each layer's full quantization parameters
    can be retrieved), and convert it into a fakely-quant model (namely, weights that are in fake-quant
    format) and fake-quant layers for the activations.
    """

    def __init__(self,
                 model: keras.models.Model,
                 is_layer_exportable_fn: Callable):

        super().__init__(model,
                         is_layer_exportable_fn)

    def export(self) -> keras.models.Model:
        """
        Convert an exportable (fully-quantized) Keras model to a fakely-quant model
        (namely, weights that are in fake-quant format) and fake-quant layers for the activations.

        Returns:
            Fake-quant Keras model.
        """

        def _unwrap_quantize_wrapper(layer: Layer):
            """
            Convert layer wrapped with ExtendedQuantizeWrapper to the layer it wraps,
             so it's ready to export. Notice that layers with activation quantization remain
             wrapped using ExtendedQuantizeWrapper to contain its activation quantizer.

            Args:
                layer: Layer to unwrap.

            Returns:
                Layer after unwrapping.

            """
            assert self.is_layer_exportable_fn(layer), f'Layer {layer.name} is not exportable.'

            # If weights are quantized, use the quantized weight for the new built layer.
            if type(layer.quantize_config) in [WeightsQuantizeConfig, WeightsActivationQuantizeConfig]:
                new_layer = layer.layer.__class__.from_config(layer.layer.get_config())
                with tf.name_scope(new_layer.name):
                    new_layer.build(layer.input_shape)

                # Build a list of the layer's new weights.
                weights_list = []
                # Go over weights, check if they should be quantized, and quantize if this is the case:
                for w in new_layer.weights:
                    val = None
                    for qw in layer.weights:
                        if w.name in qw.name:
                            # Use quantized weight if layer attribute should be quantized.
                            # For example: check if 'kernel_0' is an attribute
                            # that should be quantized. First, extract 'kernel' from variable name, check if the
                            # quantize config contains this as an attribute for quantization. If so -
                            # Take the quantized weight from the quantize_config and set it to the new layer.
                            if w.name.split('/')[-1].split(':')[0] in layer.quantize_config.get_config()['weight_attrs']:
                                val = layer.quantize_config.get_weights_and_quantizers(layer.layer)[0][1].weight
                            else:
                                val = qw
                    if val is None:
                        Logger.error(f'Could not match weight name: {w.name}')
                    weights_list.append(val)

                new_layer.set_weights(weights_list)
                new_layer.trainable = False

                # If activations are also quantized, wrap the layer back using ActivationQuantizeConfig
                # from original wrapper (weights wrapping is no longer needed).
                if isinstance(layer.quantize_config, WeightsActivationQuantizeConfig):
                    new_layer = ExtendedQuantizeWrapper(new_layer, layer.quantize_config.act_config)

                return new_layer

            # If this is a layer with activation quantization only, just return it
            # as activation quantization in the fake-quant case uses the wrapper for quantization.
            elif type(layer.quantize_config) in [ActivationQuantizeConfig]:
                return layer

            # Ideally we want in the case of no quantization to simply use the inner layer.
            # But for some reason when using SNC we are having issues to use the inner layer.
            # The clone_model method tries to reconstruct a model from the unwrapped configuration,
            # but when we have two TFOpLambda (like in the case of SNC: add and pad) one after another,
            # the output shape of the first one is in correct (it adds a new axis
            elif isinstance(layer.quantize_config, NoOpQuantizeConfig):
                return layer

        # clone each layer in the model and apply _unwrap_quantize_wrapper to layers wrapped with a QuantizeWrapper.
        self.exported_model = tf.keras.models.clone_model(self.model,
                                                          input_tensors=None,
                                                          clone_function=_unwrap_quantize_wrapper)

        return self.exported_model

    @staticmethod
    def get_custom_objects() -> Dict[str, type]:
        """

        Returns: A dictionary with objects for loading the exported model.

        """
        return {ExtendedQuantizeWrapper.__name__: ExtendedQuantizeWrapper,
                ActivationQuantizeConfig.__name__: ActivationQuantizeConfig,
                FakeQuantQuantizer.__name__: FakeQuantQuantizer,
                NoOpQuantizeConfig.__name__: NoOpQuantizeConfig}

    def save_model(self, save_model_path: str) -> None:
        """
        Save exported model to a given path.
        Args:
            save_model_path: Path to save the model.

        Returns:
            None.
        """
        if self.exported_model is None:
            Logger.critical(f'Exporter can not save model as it is not exported')
        keras.models.save_model(self.exported_model, save_model_path)



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import tensorflow as tf
from packaging import version

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if version.parse(tf.__version__) < version.parse("2.6"):
    from tensorflow.python.keras.engine.node import Node as KerasNode
    from tensorflow.keras.layers import InputLayer
    from tensorflow.python.keras.engine.functional import Functional
    from tensorflow.python.keras.engine.sequential import Sequential
else:
    from keras.engine.input_layer import InputLayer
    from keras.engine.node import Node as KerasNode
    from keras.engine.functional import Functional
    from keras.engine.sequential import Sequential

from model_compression_toolkit.core.common.graph.base_node import BaseNode


def is_node_an_input_layer(node: BaseNode) -> bool:
    """
    Checks if a node represents a Keras input layer.
    Args:
        node: Node to check if its an input layer.

    Returns:
        Whether the node represents an input layer or not.
    """
    if isinstance(node, BaseNode):
        return node.type == InputLayer
    elif isinstance(node, KerasNode):
        return isinstance(node.layer, InputLayer)
    else:
        raise Exception('Node to check has to be either a graph node or a keras node')


def is_node_a_model(node: BaseNode) -> bool:
    """
    Checks if a node represents a Keras model.
    Args:
        node: Node to check if its a Keras model by itself.

    Returns:
        Whether the node represents a Keras model or not.
    """
    if isinstance(node, BaseNode):
        return node.type in [Functional, Sequential]
    elif isinstance(node, KerasNode):
        return isinstance(node.layer, Functional) or isinstance(node.layer, Sequential)
    else:
        raise Exception('Node to check has to be either a graph node or a keras node')




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from torch import permute
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.base_node import BaseNode


class PermuteCallMethod(common.BaseSubstitution):
    """
    Find "permute" node to substitute new dimension argument if needed
    """

    def __init__(self):
        """
        Matches: 'permute' node
        """
        nodes = NodeOperationMatcher(permute)
        super().__init__(matcher_instance=nodes)

    def substitute(self,
                   graph: Graph,
                   node: BaseNode) -> Graph:
        """
        Wrap dimension of permute with tuple if it's missing

        Args:
            graph: Graph we apply the substitution on.
            node: node that match the pattern in the substitution init.

        Returns:
            Graph after applying the substitution.
        """
        # Check op_call_args is not empty and has its argument as a tuple
        if node.op_call_args and not isinstance(node.op_call_args[0], tuple):
            node.op_call_args = [node.op_call_args]
        return graph
